
/* Value path */

#include <string.h>
#include <stdlib.h>
#include <float.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>

#include "vpath.h"
#include "vector.h"
#include "numeric.h"

#define EMO_VPath_M 3
#define EMO_VPath_M2 (int) EMO_VPath_M / 2 

void EMO_VPath_alloc(EMO_VPath *vpath, int resolution, int max_size, int dim) {
  int n, theta;

  if(dim < 2) {
    printf("Error, dimension must be greather or equal than two.\n");
    exit(1);
  }

  n = ((dim - 1) * dim) / 2;
  theta = resolution * max_size;
  theta /= n;

  if(theta < 3) {
    printf("Error, not enough resolution.\n");
    printf("n %d, x %d, y %d, theta %d\n", n, vpath->x, vpath->y, vpath->theta);
    exit(1);
  }

  vpath->theta = theta;
  vpath->x = theta * n;
  vpath->y = theta;
  vpath->dim = dim;

  printf("par-objs %d, x %d, y %d, theta %d\n", n, vpath->x, vpath->y, vpath->theta);
  vpath->M = (int *) calloc(sizeof(int), vpath->x * vpath->y);
  vpath->c = (double *) calloc(sizeof(double), max_size);
}

void EMO_VPath_free(EMO_VPath *vpath) {
  free(vpath->M);
  free(vpath->c);
}

void EMO_VPath_write(EMO_VPath *vpath, const char *str) {
  int i, j;
  FILE *fp;

  if(strcmp(str, "stdout") == 0) {
    fp = stdout;
    fflush(stdout);
  }
  else {
    if((fp = fopen(str, "w")) == NULL) {
      printf("Error to open file %s\n", str);
      exit(1);
    }
  }

  if(fprintf(fp, "# %d %d\n", vpath->y, vpath->x) == 0) {
    printf("Error to write file %s\n", str);
    exit(1);
  }
 
  for(i = vpath->y-1; i > -1; i--) {
    for(j = 0; j < vpath->x; j++) {
      if(fprintf(fp, "%.2f ", (double) vpath->M[i*vpath->x+j]) == 0) {
        printf("Error to write file %s\n", str);
        exit(1);
      }
    }
    fprintf(fp, "\n");
  }

  if(strcmp(str, "stdout") != 0) 
    fclose(fp); 
}

/* M: matrix of value path
 * data: population (objectives)
 * size: number of individuals
 * ref: reference line
 * dim: number of objectives
 * theta: [1,n] resolution (times the size of the population)
 */
void EMO_VPath_run(EMO_VPath *vpath, double *data, int *filter, int size, double z) {
  int i, j, objs, k, a, b, count, flag;
  double  v1, v2, m, t, y;

  memset(vpath->M, 0, sizeof(int) * vpath->x * vpath->y);

  for(i = 0; i < size; i++) {

    flag = 0;

    if(filter != NULL && filter[i] == 0) continue;

    for(j = 0; j < vpath->dim; j++) {
      if(data[i*vpath->dim + j] > z) {
        flag = 1;
        break;
      }
    }

    if(flag)
      continue;

    objs = vpath->dim - 1;
    count = 0;

    for(j = 0; j < objs; j++) {
      for(k = j+1; k <= objs; k++) {
        v1 = data[i * vpath->dim + j];
        v2 = data[i * vpath->dim + k];

        m = (v2 - v1) / (double) (vpath->theta - 1);

        for(b = 0; b < vpath->theta; b++) {
          y = m * ((double) b) + v1;
          t = (vpath->y - 1.0) * (1.0 - y / z);
          a = round(t);

          if(a >= 0 && a < vpath->y) { 
            vpath->M[a * vpath->x + (b + count * vpath->theta) ]++;
          
          }
        }
        count++;
      }
    }
  }

  // arriba
  for(j = 0; j < vpath->x; j++) {
    i = 0;
 
    while(i < vpath->y && vpath->M[i*vpath->x + j] == 0) {
      vpath->M[i*vpath->x + j] = -1;
      i++;
    }
  }

}

// lst: list of extreme points, nadir
void EMO_VPath_contribution(EMO_VPath *vpath, double *data, int *filter, int size, double z, EMO_List *lst) {
  int objs, count, cempty, cedge, cfilled, filled, flag;
  int i, j, k, a, b, bt, c, mx, my, s1, s2, e1, e2, ry;
  double  v1, v2, m, y, t, vmax, vmin;

  vmax = 0;
  vmin = DBL_MAX;

  for(i = 0; i < size; i++) {
    if(filter != NULL && filter[i] == 0) continue;
    vpath->c[i] = 0;
    flag = 0;

    for(j = 0; j < vpath->dim; j++) {
      if(data[i*vpath->dim + j] > z) {
        if(lst == NULL || (lst != NULL && EMO_List_find(lst, i) == 0)) {
          vpath->c[i] = INT_MAX;
          flag = 1;
          break;
        }
      }

      if(data[i*vpath->dim + j] == z) {
        if(lst == NULL || (lst != NULL && EMO_List_find(lst, i) == 1)) {
          vpath->c[i] = 0;
          flag = 1;
          break;
        }
      }
    }

    if(flag)
      continue;

    objs = vpath->dim - 1;
    count = 0;

    for(j = 0; j < objs; j++) {
      for(k = j+1; k <= objs; k++) {
        v1 = data[i * vpath->dim + j];
        v2 = data[i * vpath->dim + k];
        m = (v2 - v1) / (double) (vpath->theta - 1);

        for(bt = 0; bt < vpath->theta; bt++) {
          y = m * ((double) bt) + v1;
          t = (vpath->y - 1.0) * (1.0 - y / z);
          a = round(t);
          b = bt + count * vpath->theta;

          if(a >= 0 && a < vpath->y) {
            cedge = cempty = cfilled = filled = 0;
            ry = vpath->y - a;

            s1 = a - EMO_VPath_M2;
            e1 = a + EMO_VPath_M2;
            s2 = b - EMO_VPath_M2;
            e2 = b + EMO_VPath_M2;

            for(my = s1; my <= e1; my++) {
              for(mx = s2; mx <= e2; mx++) {

                if(mx >= 0 && mx < vpath->x && my >= 0 && my < vpath->y) {
                  c = vpath->M[my*vpath->x + mx];

                  if(c == -1)
                    cedge++;
                  else if(c == 0)
                    cempty++;
                  else
                    if(!(my == a && mx == b)) {
                      filled += c;
                      cfilled++;
                    }
                }
              }
            }

            vpath->c[i] += ry * vpath->M[a*vpath->x + b];

            if(cempty > 0 && cedge > 0)  /* Individuals crossing the upper-Pareto coordinate */
              vpath->c[i] += (a + 1) * (cedge * cempty);
            else 
              vpath->c[i] -= (a + 1) * (cedge + cempty);

            if(cfilled > 0) {
              if(vpath->dim > 2 && (b % 2) == 0)
                vpath->c[i] -= ry * (cfilled);
              else
                vpath->c[i] += ry * ((double)(filled / ((double)cfilled)));
            }
          }
        }
        count++;
      }
    }

    if(vmax < vpath->c[i])
      vmax = vpath->c[i];

    if(vmin > vpath->c[i])
      vmin = vpath->c[i] - 1.0;
  }

  for(i = 0; i < size; i++) {
    if(vpath->c[i] == INT_MAX) {
      vpath->c[i] = EMO_vnorm(data + i *vpath->dim, 2.0, vpath->dim);
    }
    else if(vpath->c[i] != 0 && (vmax - vmin) != 0) {
      vpath->c[i] = (vpath->c[i] - vmin) / (vmax - vmin);
    }
  }
}


void EMO_VPath_update(EMO_VPath *vpath, double *v, double z, int val) {
  int i, j, k, a, b, bt, count, objs;
  double y, t, m;

  for(j = 0; j < vpath->dim; j++) {
    if(v[j] > z) {
      return;
    }
  }

  objs = vpath->dim - 1;
  count = 0;

  for(j = 0; j < objs; j++) {
    for(k = j+1; k <= objs; k++) {
      m = (v[k] - v[j]) / (double) (vpath->theta - 1);

      for(bt = 0; bt < vpath->theta; bt++) {
        y = m * ((double) bt) + v[j];
        t = (vpath->y - 1.0) * (1.0 - y / z);
        a = round(t);
        b = bt + count * vpath->theta;

        if(a >= 0 && a < vpath->y) {

          i = a * vpath->x + b;
          vpath->M[i] += val;

          if(val == -1 && vpath->M[i] == 0) {
            if(a == 0 || vpath->M[(a-1) * vpath->x + b] == -1) {
              vpath->M[i] = -1;

              while(++a < vpath->y && vpath->M[a * vpath->x + b] == 0)
                vpath->M[a * vpath->x + b] = -1;
            }
          }

          if(val == 1 && vpath->M[i] == 0) {
            vpath->M[i] = 1;

            while(++a < vpath->y && vpath->M[a * vpath->x + b] == -1)
              vpath->M[a * vpath->x + b] = 0;
          }
        }
      }
      count++;
    }
  }
}

void EMO_VPath_prune(EMO_VPath *vpath, double *data, int *filter, int size, int new_size, double z) {
  int i, n, j;

  n = size - new_size;
  if(n <= 0) return;

  for(i = 0; i < n; i++) {
    EMO_VPath_run(vpath, data, filter, size, z);
    EMO_VPath_contribution(vpath, data, filter, size, z, NULL);
    EMO_dmax(&j, vpath->c, filter, size);
    filter[j] = 0;
  }
}

